#ifndef HPP_HIK_SOLVER_FOR_ROBOT_H
#define HPP_HIK_SOLVER_FOR_ROBOT_H

#include "hpp/hik/solver-basic.hh"
#include "robotDynamics/jrlDynamicRobot.h"
/**
\brief This is a ChppHikSolverBasic-class object for inverse kinematics problems for humanoid robot. The humanoid robot is supposed to have at least one fixed body in the environment.
*/
class ChppHikSolver: public ChppHikSolverBasic
{
public:
    
    /**
    \brief Constructor
     */
    ChppHikSolver(CjrlDynamicRobot& inRobot);

    /** Select the joint whose body is to be considered root of robot */
    void rootJoint(CjrlJoint& inRootJoint);
    
    /**
    \brief Set active parameters by setting a mask
    \param inMask binary vector of size attRobot.currentConfiguration().size() : 0 for disabled parameter and 1 for active
     */
    bool setMask(const vectorN& inMask);

    /**
    \name Computing and retrieving the solution configuration
    @{
     */
    /**
    Compute the solution configuration
    */
    bool computeSolutionConfiguration(const vectorN& inVelocitySolution);
    /**
    \brief Get the full solution configuration including the root configuration in roll pitch yaw
    \return a vector of size robot()->numberDof()
     */
    const vectorN& solutionConfiguration();
    /**
    \brief Get the new pose of the root as a 6d vector: X Y Z ROLL PITCH YAW
    \return a pair of vectors: first vector is [X Y Z], second is [ROLL PITCH YAW]
     */
    const std::pair<vector3d,vector3d>& solutionRootConfiguration();
    /**
    \brief Get the new pose of the root as a homogenous matrix
    \return the new transformation matrix of the root
     */
    const matrix4d& solutionRootPose();
    /**
    @}
     */

private:

    CjrlDynamicRobot* attRobot;
    vectorN attBackupConfig,CurFullConfig;
    CjrlJoint*  RootJoint;
    matrix4d H0;
    matrix4d Hif;
    matrix4d Hf;
    matrix4d HRC;
    matrix4d InvHf;
    matrix3d TmpR;
    std::vector<CjrlJoint*> supportJoints;
    bool attChangeRootJoint, attChangeRootPose;
    std::pair<vector3d,vector3d> attRetPair;
    
};

#endif
