/*
 *  Copyright 2009 CNRS-LAAS
 *
 *  Authors: Oussama Kanoun
 */

#include <cstdio>
#include <unistd.h>
#include <string.h>
#include <iostream>

#include "MatrixAbstractLayer/MatrixAbstractLayer.h"
#include "dynamicsJRLJapan/dynamicsJRLJapanFactory.h"
#include "hrp2Dynamics/hrp2OptHumanoidDynamicRobot.h"
#include "hpp/gik/robot/standing-robot.hh"
#include "hpp/gik/constraint/position-constraint.hh"
#include "hpp/gik/constraint/com-constraint.hh"
#include "hpp/gik/constraint/transformation-constraint.hh"
#include "hpp/gik/tools.hh"
#include "hpp/hik/solver.hh"
#include "hpp/gik/robot/robot-motion.hh"

#define M3_IJ MAL_S3x3_MATRIX_ACCESS_I_J
#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J
#define V3_I  MAL_S3_VECTOR_ACCESS

using namespace  boost::numeric::ublas;

ChppGikStandingRobot* attStandingRobot = 0;
CjrlHumanoidDynamicRobot* attRobot = 0;
dynamicsJRLJapan::ObjectFactory jrlRobotFactory;
double attMaxVelo;
vectorN attDs,att_inv_Di_Ds, attUb, attLb ;

void test ();
void createHRP2 ( const std::string& inPath );
void updateBounds();

void createHRP2 ( const std::string& inPath )
{

    attRobot = new Chrp2OptHumanoidDynamicRobot ( &jrlRobotFactory );
    std::string linkJointFname,specificitiesFile;
    linkJointFname = inPath + "/HRP2LinkJointRank.xml";
    specificitiesFile = inPath + "/HRP2Specificities.xml";
    std::string modelname = inPath + "/HRP2.wrl";

    dynamicsJRLJapan::parseOpenHRPVRMLFile ( *attRobot, modelname,linkJointFname,specificitiesFile );

    std::string property,value;
    property="ComputeZMP"; value="true";attRobot->setProperty ( property,value );
    property="TimeStep"; value="0.005";attRobot->setProperty ( property,value );
    property="ComputeAccelerationCoM"; value="false";attRobot->setProperty ( property,value );
    property="ComputeBackwardDynamics"; value="false";attRobot->setProperty ( property,value );
    property="ComputeMomentum"; value="true";attRobot->setProperty ( property,value );
    property="ComputeAcceleration"; value="true";attRobot->setProperty ( property,value );
    property="ComputeVelocity"; value="true";attRobot->setProperty ( property,value );
    property="ComputeSkewCoM"; value="false";attRobot->setProperty ( property,value );
    property="ComputeCoM"; value="true";attRobot->setProperty ( property,value );
    
    unsigned int nDof = attRobot->numberDof();
    vectorN halfsittingConf ( nDof );

    //Half sitting
    double dInitPos[40] =
    {
        0.0, 0.0, -26.0, 50.0, -24.0, 0.0, // right leg
        0.0, 0.0, -26.0, 50.0, -24.0, 0.0, // left leg
        0.0, 0.0, // chest
        0.0, 0.0, // head
        15.0, -10.0, 0.0, -30.0, 0.0, 0.0, // right arm
        10.0, // right hand clench
        15.0,  10.0, 0.0, -30.0, 0.0, 0.0, // left arm
        10.0, // left hand clench
        -10.0, 10.0, -10.0, 10.0, -10.0, // right hand parallel mechanism
        -10.0, 10.0, -10.0, 10.0, -10.0  // left hand parallel mechanism
    };

    //waist x y z
    halfsittingConf ( 0 ) = 0.0;
    halfsittingConf ( 1 ) = 0.0;
    halfsittingConf ( 2 ) = 0.6487;
    //waist roll pitch yaw
    halfsittingConf ( 3 ) = 0.0;
    halfsittingConf ( 4 ) = 0.0;
    halfsittingConf ( 5 ) = 0.0;

    //joints
    for ( unsigned int i=6;i<nDof;i++ )
        halfsittingConf ( i ) = dInitPos[i-6]*M_PI/180;

    zero_vector<double> zeros ( attRobot->numberDof() );
    attRobot->currentConfiguration ( halfsittingConf );
    attRobot->currentVelocity ( zeros );
    attRobot->currentAcceleration ( zeros );
    attRobot->computeForwardKinematics();

    //set gaze origin and direction
    vector3d gazeDir,gazeOrigin;
    gazeDir[0] = 0;
    gazeDir[1] = cos(10*M_PI/180);
    gazeDir[2] = sin(10*M_PI/180);

    gazeOrigin[0] = 0;
    gazeOrigin[1] = 0;
    gazeOrigin[2] = -0.118;

    attRobot->gaze ( gazeDir, gazeOrigin );

    ChppGik2DShape leftFootShape, rightFootShape;

    ChppGik2DVertex vert;

    vert.x = 0.135;
    vert.y = 0.079;
    leftFootShape.vertices.push_back ( vert );
    vert.x = -0.105;
    leftFootShape.vertices.push_back ( vert );
    vert.y = -0.059;
    leftFootShape.vertices.push_back ( vert );
    vert.x = 0.135;
    leftFootShape.vertices.push_back ( vert );

    vert.x = 0.135;
    vert.y = 0.059;
    rightFootShape.vertices.push_back ( vert );
    vert.x = -0.105;
    rightFootShape.vertices.push_back ( vert );
    vert.y = -0.079;
    rightFootShape.vertices.push_back ( vert );
    vert.x = 0.135;
    rightFootShape.vertices.push_back ( vert );

    attStandingRobot = new ChppGikStandingRobot ( *attRobot, leftFootShape, rightFootShape );
    attStandingRobot->staticState ( halfsittingConf );
}


void test()
{
    ChppRobotMotion attSolutionMotion ( attRobot,0.0, 5e-3);

    ChppHikSolver hikSolver ( *attRobot );
    hikSolver.rootJoint ( * ( attRobot->rightAnkle() ) );


    std::vector<CjrlGikStateConstraint*> tasks;
    vector3d p, localPoint;

    vector3d com = attRobot->positionCenterOfMass();
    ChppGikComConstraint comc ( *attRobot,com[0],com[1] );
    tasks.push_back ( &comc );

    matrix4d LFT=  attRobot->leftAnkle()->currentTransformation();
    CjrlJoint& lfJ = * ( attRobot->leftAnkle() );
    p[0] = p[1] = p[2] = 0;
    ChppGikTransformationConstraint footc ( *attRobot,lfJ,p,LFT );
    tasks.push_back ( &footc );

    matrix4d curT=  attRobot->rightWrist()->currentTransformation();
    CjrlJoint& rwJoint = * ( attRobot->rightWrist() );
    attRobot->rightHand()->getCenter(localPoint);
    p = curT*localPoint;
    ChppGikPositionConstraint pc ( *attRobot,rwJoint,localPoint, p );
    tasks.push_back ( &pc );

    vectorN solutionConfig, veloSolution ( attRobot->numberDof() );
    int retV;

    for ( unsigned int b=0;b<attUb.size();b++ )
    {
        hikSolver.defaultUpperBound ( b, attMaxVelo );
        hikSolver.defaultLowerBound ( b, -attMaxVelo );
    }

    ChppHikLinearSystem linSys(attRobot->numberDof());
    
    vectorN outEvaluation;
    vectorN mask = attStandingRobot->maskFactory()->wholeBodyMask();
    for ( unsigned int j = 0; j< 6;j++ )
        mask ( j ) = 0.0;
    hikSolver.setMask ( mask );

    ChppRobotMotionSample motionsample;

    for ( unsigned int j = 0; j< 500;j++ )
    {
        p = pc.worldTarget();
        p[0] += 0.001;
        pc.worldTarget ( p );

        updateBounds();
        hikSolver.clearBounds();
        for ( unsigned int b=6;b<attUb.size();b++ )
        {
            hikSolver.upperBound ( b, attUb ( b ) );
            hikSolver.lowerBound ( b, attLb ( b ) );
        }

        hikSolver.initFeasibleSet();

        veloSolution.clear();
        for ( unsigned int t = 0; t<tasks.size();t++ )
        {
            tasks[t]->jacobianRoot ( * ( attRobot->rightAnkle() ) );
            tasks[t]->computeValue();
            tasks[t]->computeJacobian();
        }

        for ( unsigned int t = 0; t<tasks.size();t++ )
        {

            hikSolver.clearTasks();
            linSys.matrix() = tasks[t]->jacobian();
            linSys.vector() = -tasks[t]->value();
            hikSolver.addTask ( linSys,true );
            retV = hikSolver.solve ( veloSolution );
            if ( retV!=0 )
            {
                std::cout << "Could not solve Hik problem at iteration "<<j<<" Error code = " << retV<<"\n";
                attSolutionMotion.dumpTo ( "hikmotion" );
                return;
            }
            veloSolution = hikSolver.solution();
            if ( t != tasks.size()-1 )
                hikSolver.updateFeasibleSet();
        }

        hikSolver.computeSolutionConfiguration ( veloSolution );
        solutionConfig = hikSolver.solutionConfiguration();
        attRobot->currentConfiguration ( solutionConfig );
        attRobot->computeForwardKinematics();
        motionsample.configuration = attRobot->currentConfiguration();
	 attSolutionMotion.appendSample(motionsample);

    }
    attSolutionMotion.dumpTo ( "hikmotion" );
}

void updateBounds()
{
    double ul,ll,dp,dn;
    for ( unsigned int i =0; i<attDs.size();i++ )
    {
        ul = attRobot->upperBoundDof ( i );
        ll = attRobot->lowerBoundDof ( i);
        dp = ul - attRobot->currentConfiguration() ( i );
        attUb ( i ) = attMaxVelo* ( dp-attDs ( i ) ) *att_inv_Di_Ds ( i );
        dn = attRobot->currentConfiguration() ( i )-ll;
        attLb ( i ) = -attMaxVelo* ( dn-attDs ( i ) ) *att_inv_Di_Ds ( i );
    }
}

int main ( int argc, char** argv )
{
    if ( argc !=2 )
    {
        std::cout << "Usage: simple-test [Path to folder containing files HRP2.wrl, HRP2LinkJointRank.xml and HRP2Specificities.xml]" <<std::endl;
        return  -1;
    }

    createHRP2 ( std::string ( argv[1] ) );
    //prepare for dynamic velocity bounds
    attDs.resize ( attRobot->numberDof() );
    attUb.resize ( attRobot->numberDof() );
    attLb.resize ( attRobot->numberDof() );
    att_inv_Di_Ds.resize ( attRobot->numberDof() );
    attMaxVelo = 0.002;
    double DiRatio = 0.10;
    double DiDsRatio = 0.05;
    double ul,ll,di;
    for ( unsigned int i =0; i<attDs.size();i++ )
    {
        ul = attRobot->upperBoundDof ( i  );
        ll = attRobot->lowerBoundDof ( i );
        di = DiRatio* ( ul-ll );
        attDs ( i ) = DiDsRatio*di;
        att_inv_Di_Ds ( i ) = 1/ ( di-attDs ( i ) );
    }

    test();

    delete attStandingRobot;

    return 0;
}

